from bqsqoop.utils import typed
from bqsqoop.utils.gcloud import auth, storage, bigquery


class BigqueryParquetLoadJob():
    """Loads local parquet files data into bigquery using google storage

    Init will _errors.append(a error if config is invalid)
    If invalid, errors property will contains the config errors
    and a flag `is_config_valid` is updated accordingly

    Args:
        configs (dict): Job's Bigquery sub-section configs as a dict
    """

    def __init__(self, configs):
        self._project_id = configs.get("project_id")
        self._dataset_name = configs.get("dataset_name")
        self._table_name = configs.get("table_name")
        self._gcs_tmp_path = configs.get("gcs_tmp_path")
        self._service_account_key = configs.get('service_account_key')
        self._write_truncate = configs.get('write_truncate', True)
        self._validate_configs()

    def _validate_configs(self):
        self.errors = {}
        self.is_config_valid = True
        _strings = ["project_id", "dataset_name",
                    "table_name", "gcs_tmp_path"]
        for _str_vars in _strings:
            _res = typed.non_empty_string(getattr(self, "_" + _str_vars))
            if _res:
                self.errors[_str_vars] = _res
        _res = auth.setup_credentials(self._service_account_key)
        if _res:
            self.errors["google_auth"] = _res
        if self.errors:
            self.is_config_valid = False

    def execute(self, files):
        """Executes the job to load parquet files to Bigquery tables

        Args:
            files (list_of_str): List of files to be uploaded to bigquery.
                Should be full file paths

        Returns:
            None if the job is successful, errors if failed
        """
        _gcs_dest_path = storage.copy_files_to_gcs(files, self._gcs_tmp_path,
                                                   self._project_id,
                                                   use_new_tmp_folder=True)
        bigquery.load_parquet_files(
            _gcs_dest_path + "*.parq",
            self._project_id,
            self._dataset_name,
            self._table_name,
            write_truncate=self._write_truncate
        )
        storage.delete_files_in(_gcs_dest_path, self._project_id)



import os
import uuid
import time
import logging
import traceback
import pandas as pd

import sqlalchemy
from bqsqoop.utils import parquet_util


def get_results_cursor(sql_bind, query, pool_timeout=300):
    engine = sqlalchemy.create_engine(sql_bind, pool_timeout=pool_timeout)
    connection = engine.connect()
    return connection.execute(query)


def export_to_parquet(worker_id, sql_bind, query, filter_field, start_pos,
                      end_pos, output_folder, progress_bar=True,
                      fetch_size=1000):
    try:
        start_time = int(time.time())
        output_file = os.path.join(output_folder, "{}.parq".format(
            str(uuid.uuid4())[:8]))
        if start_pos and end_pos:
            filter_query = "{0} >= {1} AND {0} <= {2}".format(
                filter_field, start_pos, end_pos)
            query = query % filter_query
        parquetUtil = parquet_util.ParquetUtil(output_file)
        engine = sqlalchemy.create_engine(sql_bind, pool_timeout=300)
        connection = engine.connect()
        for chunk in pd.read_sql_query(
            sql=query,
            con=connection,
            chunksize=fetch_size
        ):
            parquetUtil.append_df_to_parquet(chunk)
        parquetUtil.close()
        logging.info(
            "Dump -> worker_id: {0}, time_taken: {1} secs".format(
                worker_id, int(time.time()) - start_time))
        return output_file
    except Exception:
        logging.error(traceback.format_exc())
        raise



from .iextractor import Extractor   # noqa
from .elasticsearch import ElasticSearchExtractor
from .sql import SQLExtractor


_extractors = {
    "es": ElasticSearchExtractor,
    "elasticsearch": ElasticSearchExtractor,
    "sql": SQLExtractor,
}


def get_extractor_for(config):
    """Fetches extractor for given job config.

    :param config: A dict containg sqoop job config.
    returns: An Extractor object
    """
    _extractor_config = config.get("extractor")
    if(len(_extractor_config) != 1):
        raise Exception(
            "Job takes only one extractor, given {}".format(
                _extractor_config))
    _extractor_src, _extractor_config = next(iter(_extractor_config.items()))
    _extractor = _extractors.get(_extractor_src)
    if(_extractor):
        return _extractor(_extractor_config)
    else:
        raise Exception(
            "Unknown extractor given {}".format(
                _extractor_config))



import os
import uuid
import elasticsearch
import pandas as pd

from bqsqoop.utils import parquet_util
from bqsqoop.utils.progressbar_util import ProgressBar


class ESHelper():
    @classmethod
    def get_fields(self, es_hosts, index):
        _es = self._get_es_client(es_hosts)
        _mappings = _es.indices.get_mapping(
            index=index
        )
        _fields = {}
        for index, value in _mappings.items():
            index_mappings = value['mappings']
            for _, index_type_value in index_mappings.items():
                _properties = index_type_value['properties']
                _keys = list(_properties.keys())
                for _key in _keys:
                    _fields[_key] = _properties[_key]["type"]
                break
        return _fields

    @classmethod
    def scroll_and_extract_data(self, worker_id, total_worker_count, es_hosts,
                                es_timeout, search_args, fields,
                                output_folder, progress_bar=True):
        _es = self._get_es_client(es_hosts)
        search_args = self._add_slice_if_needed(
            total_worker_count, search_args, worker_id)
        _output_file = self._output_file_for(
            output_folder, search_args['index'])
        _page = _es.search(**search_args)
        _data, _sid = self._get_data_from_es_page(_page)
        _total_hits = self._get_total_hits(_page)
        df = None
        _parquetUtil = parquet_util.ParquetUtil(_output_file)
        _pbar = ProgressBar(
            total_length=_total_hits, position=worker_id, enabled=progress_bar)
        if _data:
            df = pd.DataFrame.from_dict(_data)
            self._fix_types_from_es(df, fields)
            _parquetUtil.append_df_to_parquet(df)
            _pbar.move_progress(len(_data))
            while True:
                _page = _es.scroll(scroll_id=_sid, scroll=es_timeout)
                _data, _sid = self._get_data_from_es_page(_page)
                if _data:
                    df = pd.DataFrame.from_dict(_data)
                    self._fix_types_from_es(df, fields)
                    _parquetUtil.append_df_to_parquet(df)
                    _pbar.move_progress(len(_data))
                else:
                    break
        _parquetUtil.close()
        return _output_file

    @classmethod
    def _fix_types_from_es(self, df, fields):
        for _name, _type in fields.items():
            if _type == "date":
                df[_name] = pd.to_datetime(
                    df[_name], format="%Y-%m-%dT%H:%M:%S")

    @classmethod
    def _output_file_for(self, output_folder, index):
        return os.path.join(output_folder, '{}_{}.parq'.format(
            index, str(uuid.uuid4())[:8]))

    @classmethod
    def _get_es_client(self, es_hosts):
        return elasticsearch.Elasticsearch(es_hosts)

    @classmethod
    def _add_slice_if_needed(self, total_worker_count, search_args, worker_id):
        if total_worker_count > 1:
            # Add ES scroll slicing to handle parallel scrolling of the data
            search_args['body']['slice'] = {
                'id': worker_id,
                'max': total_worker_count
            }
        return search_args

    @classmethod
    def _get_data_from_es_page(self, _page):
        if 'hits' in _page and 'hits' in _page['hits']:
            _data = _page['hits']['hits']
            if _data:
                _rows = [_datumn['_source'] for _datumn in _data]
                _sid = _page['_scroll_id']
                return _rows, _sid
        return None, None

    @classmethod
    def _get_total_hits(self, _page):
        if 'hits' in _page and 'total' in _page['hits']:
            return _page['hits']['total']
        return 0



import logging


def setup(debug_mode):
    """Streams logs to stdout

    Args:
        debug_mode (bool): a boolean to enable verbose logs
    """
    logFormatter = logging.Formatter(
        "%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
    rootLogger = logging.getLogger()

    consoleHandler = logging.StreamHandler()
    consoleHandler.setFormatter(logFormatter)
    rootLogger.addHandler(consoleHandler)

    if(debug_mode):
        rootLogger = logging.getLogger()
        rootLogger.setLevel(logging.DEBUG)
        logging.debug("Verbose logging enabled")



import re
import logging

from google.cloud import bigquery


def load_parquet_files(gcs_path, gcs_project, dataset_name, table_name,
                       write_truncate=True):
    """Creates/Updates BQ table with parquet files from GCS bucket path

    Args:
        gcs_path (str): Should be a path in GCS. Starts with gs://
        gcs_project (str): Google project id for Biguery's
        dataset_name (str): Bigquery dataset name, must exist in project
        table_name (str): Bigquery table name
        write_truncate (bool): flag to define either to Override or Append
            data to the table. Default is True.

    Returns:
        None
    """
    _validate_gcs_path(gcs_path)
    _load_files_from_gcs(gcs_path, gcs_project, dataset_name,
                         table_name, "PARQUET", write_truncate)


def _load_files_from_gcs(gcs_path, gcs_project, dataset_name,
                         table_name, file_format, write_truncate):
    logging.debug("Loading data from {0} to bq table {1}.{2}.{3}\n".format(
        gcs_path, gcs_project, dataset_name, table_name))
    _client = bigquery.Client(project=gcs_project)
    _job_config = bigquery.LoadJobConfig()
    _job_config.source_format = file_format
    _job_config.write_disposition = 'WRITE_TRUNCATE' if \
        write_truncate else 'WRITE_APPEND'
    target_table = _client.dataset(dataset_name).table(table_name)
    job = _client.load_table_from_uri(gcs_path, target_table,
                                      job_config=_job_config)
    logging.debug("Waiting for load job to finish...")
    job.result()
    logging.info("Table {} created".format(table_name))


def _validate_gcs_path(gcs_path):
    gcs_bucket_path_re_pattern = r'gs://.+'
    _match_obj = re.match(gcs_bucket_path_re_pattern, gcs_path)
    if not _match_obj:
        raise Exception("Invalid GCS path")



from abc import ABC, abstractmethod


class Extractor(ABC):
    """Extractor interface for source table extraction

    Implement the abstract methods `validate_config` and
    `execute` to implement a new type of Extractor.

    Args:
        _config: Extractor sub-config of the Sqoop Job
    """

    def __init__(self, _config):
        self._config = _config
        super().__init__()

    @abstractmethod
    def validate_config(self):
        """implement your config validator here

        Returns: (bool)
            True if valid, else False
        """
        return False    # pragma: no cover

    @abstractmethod
    def extract_to_parquet(self):
        """Extracts data from source to parquet files

        Returns:
            List of all extracted full file path.
        """
        return []   # pragma: no cover




class MissingConfigError(ValueError):
    def __init__(self, config_name, root_name):
        self.error = 'Missing config for {0} under {1}'.format(
            config_name, root_name)

    def __str__(self):
        return self.error


class MissingDataError(ValueError):
    def __init__(self):
        self.error = 'Missing data'

    def __str__(self):
        return self.error


class InvalidTypeError(ValueError):
    def __init__(self, data_type):
        self.error = "Wrong type, it's supposed to be a {0}".format(data_type)

    def __str__(self):
        return self.error



import logging
import uuid

from bqsqoop.utils import async_worker
from bqsqoop.extractor import Extractor
from bqsqoop.utils.errors import MissingConfigError
from bqsqoop.extractor.elasticsearch import helper


class ElasticSearchExtractor(Extractor):
    """Extractor data from Elasticsearch index
    """

    def __init__(self, _config):
        super().__init__(_config)
        self._no_of_workers = self._config.get('no_of_workers', 1)
        self._timeout = self._config.get('timeout', '60s')
        self._scroll_size = self._config.get('scroll_size', 1000)
        self._fields = self._config.get('fields', ['_all'])
        self._output_folder = self._config.get(
            'output_folder', "./" + str(uuid.uuid4())[:8])

    def validate_config(self):
        """Validates required configs for Elasticsearch extraction
        """
        _es_root_name = 'Elasticsearch'
        if 'url' not in self._config:
            raise MissingConfigError('url', _es_root_name)
        if 'index' not in self._config:
            raise MissingConfigError('index', _es_root_name)
        return None

    def extract_to_parquet(self):
        _async_worker = async_worker.AsyncWorker(
            self._no_of_workers)
        for i in range(self._no_of_workers):
            _async_worker.send_data_to_worker(
                worker_id=i,
                **self._get_extract_job_fn_and_params()
            )
        logging.debug('Waiting for Extractor job results...')
        return _async_worker.get_job_results()

    def _get_extract_job_fn_and_params(self):
        search_args = dict(
            index=self._config['index'],
            scroll=self._timeout,
            size=self._scroll_size,
            body={
                'query': {
                    'match_all': {}
                }
            }
        )
        if '_all' not in self._fields:
            search_args['_source_include'] = ','.join(
                self._fields)
        _fields = helper.ESHelper.get_fields(
            self._config['url'], self._config['index'])
        return dict(worker_callback=helper.ESHelper.scroll_and_extract_data,
                    total_worker_count=self._no_of_workers,
                    es_hosts=self._config['url'],
                    es_timeout=self._timeout,
                    output_folder=self._output_folder,
                    search_args=search_args,
                    fields=_fields)



import logging

from bqsqoop.utils import log
from bqsqoop import extractor
from bqsqoop.utils.gcloud import job as bq_job


class Job():
    """A bq-sqoop's Job definition

    Given a set of configs will create a Sqoop job to extract data from
    source and load it to Google Bigquery table.

    example:

    Args:
        configs (dict): Configs for the job execution
            More examples on config definition at example/configs
        debug_mode: A Optional boolean representing log level,
            Default is False
    """

    def __init__(self, configs, debug_mode=False):
        log.setup(debug_mode)
        self._configs = configs
        self._validate_configs()
        logging.debug("Loading the job with configs: {}".format(self._configs))

    def _validate_configs(self):
        self._extractor = extractor.get_extractor_for(self._configs)
        self._extractor.validate_config()
        self._validate_bq_configs()

    def _validate_bq_configs(self):
        _bq_config = self._configs.get("bigquery")
        if not _bq_config:
            raise Exception("Missing bigquery configs")
        self._bq_job = bq_job.BigqueryParquetLoadJob(_bq_config)
        if not self._bq_job.is_config_valid:
            logging.error("Invalid Bigquery configs: {}".format(
                self._bq_job.errors))
            raise Exception("Invalid Bigquery configs")

    def execute(self):
        """Executes the job of extracting data to  Bigquery
        """
        _extracted_files = self._extractor.extract_to_parquet()
        self._bq_job.execute(_extracted_files)



import math

from datetime import datetime, timedelta


def _calculate_splits_for_int(min_value, max_value, n):
    count_per_batch = int(
        math.ceil((max_value - min_value + 1) / (n * 1.0)))
    splits = []
    for i in range(n):
        splits.append({
            'start': (count_per_batch * (i) + min_value),
            'end': ((count_per_batch * (i) + min_value) + count_per_batch - 1)
        })
    return splits


def _calculate_splits_for_timestamp(min_value, max_value, n):
    diff_secs = (max_value - min_value).total_seconds()
    count_per_batch = int(math.ceil((diff_secs + 1) / (n * 1.0)))
    splits = []
    for i in range(n):
        splits.append({
            'start': (min_value + timedelta(seconds=(count_per_batch * i))),
            'end': (
                (timedelta(seconds=(count_per_batch * i)) + min_value) +
                timedelta(seconds=(count_per_batch)))
        })
    return splits


_split_fn_map = {
    int: _calculate_splits_for_int,
    datetime: _calculate_splits_for_timestamp
}


def calculate_splits(min_value, max_value, n):
    """Splits and returns n ranges between min_value and max_value

    Supports both int and timestamp type values.

    """
    fn = _split_fn_map.get(type(min_value))
    if fn:
        return fn(min_value, max_value, n)
    else:
        raise Exception("Type of the split-by column is not supported.")



import logging
import uuid

from bqsqoop.utils import async_worker, math_util
from bqsqoop.extractor import Extractor
from bqsqoop.utils.errors import MissingConfigError
from bqsqoop.extractor.sql import helper


class SQLExtractor(Extractor):
    """Extractor data from sql based DB

    Any DB supported by sqlalchemy should work

    """

    def __init__(self, _config):
        super().__init__(_config)
        self._no_of_workers = self._config.get('no_of_workers', 1)
        self._output_folder = self._config.get(
            'output_folder', "./" + str(uuid.uuid4())[:8])

    def validate_config(self):
        """Validates required configs for SQL extraction
        """
        root_name = 'SQL'
        if 'sql_bind' not in self._config:
            raise MissingConfigError('sql_bind', root_name)
        if 'query' not in self._config:
            raise MissingConfigError('query', root_name)
        if self._no_of_workers > 1 and 'filter_field' not in self._config:
            raise MissingConfigError('filter_field', root_name)
        return None

    def extract_to_parquet(self):
        if self._no_of_workers > 1:
            _async_worker = async_worker.AsyncWorker(
                self._no_of_workers)
            splits = self._add_filter_to_query()
            for i in range(self._no_of_workers):
                _async_worker.send_data_to_worker(
                    worker_id=i,
                    **self._get_extract_job_fn_and_params(splits[i])
                )
            logging.debug('Waiting for Extractor job results...')
            return _async_worker.get_job_results()
        else:
            args = self._get_extract_job_fn_and_params(None)
            del args['worker_callback']
            res = helper.export_to_parquet(
                worker_id=0,
                **args
            )
            return [res]

    def _add_filter_to_query(self):
        min_id, max_id = self._get_min_max()
        splits = math_util.calculate_splits(
            min_id, max_id, self._no_of_workers)
        return splits

    def _get_min_max(self):
        filter_field = self._config['filter_field']
        query = self._config['query']
        sql_bind = self._config['sql_bind']
        sub_mix_max_query = query % "1=1"
        sql_min_max_query = """
            select min(t.{0}), max(t.{0}) from ({1}) as t
        """.format(filter_field, sub_mix_max_query)
        logging.debug("Running PSQL query: {}".format(sql_min_max_query))
        min_id, max_id = helper.get_results_cursor(
            sql_bind, sql_min_max_query).fetchone()
        return min_id, max_id

    def _get_extract_job_fn_and_params(self, split_range):
        start_pos = end_pos = None
        if split_range:
            start_pos = split_range['start']
            end_pos = split_range['end']
        return dict(worker_callback=helper.export_to_parquet,
                    sql_bind=self._config['sql_bind'],
                    query=self._config['query'],
                    filter_field=self._config.get('filter_field'),
                    output_folder=self._output_folder,
                    start_pos=start_pos,
                    end_pos=end_pos)



import re
import uuid
import logging

from google.cloud import storage


def copy_files_to_gcs(files, gcs_bucket_path, project_id,
                      use_new_tmp_folder=False):
    """copies given files to GCS bucket path

    Args:
        files (list_of_str): List of files to be uploaded.
            Should be full file paths
        gcs_bucket_path (str): Should be a path in GCS.
            Starts with gs://
        use_new_tmp_folder (bool): If set true will upload all files
            under a new folder in gcs_bucket_path

    Returns:
        str: path to the folder in gcs where the files are uploaded
    """
    _validate_gcs_path(gcs_bucket_path)
    client = storage.Client(project=project_id)
    bucket_name, folder_path, _ = _get_details_from_gcs_path(
        gcs_bucket_path, use_new_tmp_folder)
    bucket = client.get_bucket(bucket_name)
    for file in files:
        filename = file.split('/')[-1]
        blob = bucket.blob(folder_path + filename)
        blob.upload_from_filename(filename=file)
    return "gs://" + bucket_name + "/" + folder_path


def delete_files_in(gcs_bucket_path, project_id, delimiter=None):
    """Deletes all files in given GCS bucket path

    Args:
        gcs_bucket_path (str): Should be a path in GCS.
            Starts with gs://
        delimiter (str): can be used to restrict the results to only the
            "files" in the given "folder". Without the delimiter,
            the entire tree under the prefix is returned.
    Returns:
        None
    """
    _validate_gcs_path(gcs_bucket_path)
    client = storage.Client(project=project_id)
    bucket_name, folder_path, _ = _get_details_from_gcs_path(
        gcs_bucket_path, False)
    bucket = client.get_bucket(bucket_name)
    blobs = bucket.list_blobs(prefix=folder_path, delimiter=delimiter)
    for blob in blobs:
        blob.delete()
        logging.debug("Deleted blob: {}".format(blob.name))


def download_file_as_string(gcs_uri):
    """Download a file from given GCS uri

    Args:
        gcs_uri (str): Should be a path in GCS.
            Starts with gs://

    Returns: (str)
        Files content as a string
    """
    _validate_gcs_path(gcs_uri)
    client = storage.Client()
    bucket_name, _, file_path = _get_details_from_gcs_path(
        gcs_uri, False)
    bucket = client.get_bucket(bucket_name)
    blob = bucket.get_blob(file_path)
    return blob.download_as_string()


def _validate_gcs_path(gcs_path):
    gcs_bucket_path_re_pattern = r'gs://.+'
    _match_obj = re.match(gcs_bucket_path_re_pattern, gcs_path)
    if not _match_obj:
        raise Exception("Not a valid GCS tmp path.")


def _get_details_from_gcs_path(gcs_bucket_path, use_new_tmp_folder):
    gcs_bucket_path = gcs_bucket_path.replace("gs://", "")
    bucket_name = gcs_bucket_path.split("/")[0]
    file_path = "/".join(gcs_bucket_path.split("/")[1:])
    folder_path = file_path
    if folder_path != "" and not re.match(r'(.*)\/$', folder_path):
        folder_path = folder_path + '/'
    if use_new_tmp_folder:
        folder_path = folder_path + str(uuid.uuid4()) + "/"
    return bucket_name, folder_path, file_path



import tqdm


class ProgressBar():
    def __init__(self, total_length, position=0, enabled=True):
        """Helper for a progress bar in cli

        Args:
            total_length (int): 100% length of the progress bar
            position (int): Initial postition to start the progress bar at
                Default: 0
            enabled (bool): To make a calls to helper but not actually show
                a progress bar. Default: True
        """
        self._enabled = enabled
        if self._enabled:
            self._progress_bar = tqdm.tqdm(
                total=total_length, position=position,
                desc="slice-%s" % position, ascii=True)

    def move_progress(self, length):
        """Move the progress bar to n position

        Args:
            length (int): N positions to move w.r.t total_length
        """
        if self._enabled:
            self._progress_bar.update(length)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import os
import re
import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


def version():
    import bqsqoop
    return bqsqoop.__version__


README = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()

egg = re.compile("\#egg\=(.*?)$")
requirements = open('requirements.txt').read().splitlines()
REQUIREMENTS = [req for req in requirements if (
    not req.startswith('-e') and not req.startswith('#') and req != '')]
DEPENDENCY_LINKS = [
    req.replace('-e ', '') for req in requirements if req.startswith('-e')
]
REQUIREMENTS.extend(
    [egg.findall(req)[0] for req in requirements if req.startswith('-e')]
)
REQUIREMENTS.append('futures; python_version == "2.7"')


test_requirements = open('test-requirements.txt').read().splitlines()
TEST_REQUIREMENTS = [req for req in test_requirements if (
    not req.startswith('-e') and not req.startswith('#') and req != '')]

setup(
    name='bq-sqoop',
    author='Arun Kumar Ramanathan',
    author_email='therealrako@gmail.com',
    license='MIT License',
    url='https://github.com/therako/bqsqoop',
    description='A CLI client for exporting elasticsearch data to csv',
    long_description=README,
    version=version(),
    packages=find_packages(exclude=('tests',)),
    cmdclass={'test': PyTest},
    scripts=('bin/bq-sqoop',),
    install_requires=REQUIREMENTS,
    tests_require=TEST_REQUIREMENTS,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Development Status :: 2 - Pre-Alpha',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
)



import logging

from concurrent.futures import ProcessPoolExecutor, wait


class AsyncWorker(object):
    """
        Wrapper for ProcessPoolExectuor

        Args:
            num_workers - Number of parallel workers to run
        Note: other functions `send_data_to_worker` and `get_job_results`
    """

    def __init__(self, num_workers, num_jobs=None):
        self.num_workers = num_workers
        self._executor = ProcessPoolExecutor(max_workers=self.num_workers)
        self._futures = []

    def send_data_to_worker(self, worker_callback, **args):
        """
            Enqueue's a python callable to the async executor

            Args:
                worker_callback - A python callable method
                **args - args for worker_callback

        """
        self._futures.append(self._executor.submit(worker_callback, **args))

    def get_job_results(self):
        """
            Waits and gets results from all jobs

            Returns:
                array of return values from each task
            Note: It checks for exceptions inside the tasks and in case of
            any exceptions, they will be printed
            and a final RuntimeError will be raisen.
        """
        _job_results = []
        raise_error = False
        all_futures = wait(self._futures).done
        for x in all_futures:
            try:
                _job_results.append((x.result()))
            except Exception as exc:
                logging.error('A slice ended with Exception: %s' % exc)
                raise_error = True
        if raise_error:
            raise RuntimeError("One or more worker ended in Exception.")
        return _job_results



import os


def setup_credentials(key=None):
    """Setups up GCP credentials

    If key's given will setup gcloud utils configs from it,
    else will check for existence default credentials to use

    Args:
        key (str): Google service account key json string

    Returns:
        error if any, else None
    """
    if key:
        return _setup_credentials_from_config(key)
    # TODO: Properly Validate the deafault serivce account here
    # else:
    #     return _check_for_default_credential()


def _check_for_default_credential():
    _credential_env = os.getenv('GOOGLE_APPLICATION_CREDENTIALS', None)
    if not _credential_env:
        return Exception(
            "GCP credentials not found in the VM, " +
            "Please set GOOGLE_APPLICATION_CREDENTIALS in global env")


def _setup_credentials_from_config(key):
    _service_account_key_file = '/tmp/gcp_service_account_key.json'
    with open(_service_account_key_file, 'w+') as f:
        f.write(key)
    os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = _service_account_key_file



import pyarrow as pa
import pyarrow.parquet as pq


class ParquetUtil():
    def __init__(self, output_file):
        """Helper for writing Parquet files

        Args:
            output_file (str): output file name with full path
        """
        self._output_file = output_file
        self._pqwriter = None

    def write_df_to_parquet(self, df, preserve_index=False, close_writer=True):
        """Writes Pandas Dataframe to a new Parquet file

        Closes the writer after writing all of dataframe and before returning

        Args:
            df (Pandas Datafram): Data to be written to parquet file.
            preserve_index (bool): Set this to True if you want to
                preserve the index of pandas dataframe, default is False
                and indexes are dropped.
            close_writer (bool): Set to true closes the writer at the end
                Default: True
        """
        table = pa.Table.from_pandas(df, preserve_index=preserve_index)
        self._pqwriter = pq.ParquetWriter(self._output_file, table.schema)
        self._pqwriter.write_table(table)
        if close_writer:
            self.close()

    def append_df_to_parquet(self, df, preserve_index=False):
        """Writes Pandas Dataframe to a new Parquet file in Append mode

        Can be used to append data to the same parquet file multiple times,
        need to call close() funciton after done with all appends.

        Args:
            df (Pandas Datafram): Data to be written to parquet file.
            preserve_index (bool): Set this to True if you want to
                preserve the index of pandas dataframe, default is False
                and indexes are dropped.
        """
        if not self._pqwriter:
            self.write_df_to_parquet(df, preserve_index=preserve_index,
                                     close_writer=False)
        else:
            table = pa.Table.from_pandas(
                df, preserve_index=preserve_index)
            self._pqwriter.write_table(table)

    def close(self):
        """Closes the parquet writer to the output file

        It's safe to call it multiple times, will only close if a writer
        is open.
        """
        if self._pqwriter:
            self._pqwriter.close()
            self._pqwriter = None



"""Typed module has useful functions to validate types

examples:

"""
from bqsqoop.utils.errors import (
    MissingDataError, InvalidTypeError
)


def non_empty_string(data):
    if data is None or data == "":
        return MissingDataError()
    elif type(data) != str:
        return InvalidTypeError("string")
